# coding: utf-8

# flake8: noqa
"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from client.models.alert_condition_metric_type import AlertConditionMetricType
from client.models.alert_condition_severity import AlertConditionSeverity
from client.models.autoscaling_policy import AutoscalingPolicy
from client.models.binary_classification_output import BinaryClassificationOutput
from client.models.config import Config
from client.models.container import Container
from client.models.custom_predictor import CustomPredictor
from client.models.deployment_mode import DeploymentMode
from client.models.endpoint_status import EndpointStatus
from client.models.env_var import EnvVar
from client.models.environment import Environment
from client.models.file_format import FileFormat
from client.models.gpu_config import GPUConfig
from client.models.gpu_toleration import GPUToleration
from client.models.label import Label
from client.models.logger import Logger
from client.models.logger_config import LoggerConfig
from client.models.logger_mode import LoggerMode
from client.models.metrics_type import MetricsType
from client.models.mock_response import MockResponse
from client.models.model import Model
from client.models.model_endpoint import ModelEndpoint
from client.models.model_endpoint_alert import ModelEndpointAlert
from client.models.model_endpoint_alert_condition import ModelEndpointAlertCondition
from client.models.model_endpoint_rule import ModelEndpointRule
from client.models.model_endpoint_rule_destination import ModelEndpointRuleDestination
from client.models.model_prediction_config import ModelPredictionConfig
from client.models.model_prediction_output import ModelPredictionOutput
from client.models.model_prediction_output_class import ModelPredictionOutputClass
from client.models.model_schema import ModelSchema
from client.models.operation_tracing import OperationTracing
from client.models.pipeline_tracing import PipelineTracing
from client.models.prediction_job import PredictionJob
from client.models.prediction_job_config import PredictionJobConfig
from client.models.prediction_job_config_bigquery_sink import PredictionJobConfigBigquerySink
from client.models.prediction_job_config_bigquery_source import PredictionJobConfigBigquerySource
from client.models.prediction_job_config_gcs_sink import PredictionJobConfigGcsSink
from client.models.prediction_job_config_gcs_source import PredictionJobConfigGcsSource
from client.models.prediction_job_config_model import PredictionJobConfigModel
from client.models.prediction_job_config_model_result import PredictionJobConfigModelResult
from client.models.prediction_job_resource_request import PredictionJobResourceRequest
from client.models.prediction_logger_config import PredictionLoggerConfig
from client.models.project import Project
from client.models.protocol import Protocol
from client.models.ranking_output import RankingOutput
from client.models.regression_output import RegressionOutput
from client.models.resource_request import ResourceRequest
from client.models.result_type import ResultType
from client.models.save_mode import SaveMode
from client.models.schema_spec import SchemaSpec
from client.models.secret import Secret
from client.models.standard_transformer_simulation_request import StandardTransformerSimulationRequest
from client.models.standard_transformer_simulation_response import StandardTransformerSimulationResponse
from client.models.transformer import Transformer
from client.models.value_type import ValueType
from client.models.version import Version
from client.models.version_endpoint import VersionEndpoint
